

//From Numerical Recipes in C, with personal comments added.
/*
 *   Replaces data[1..2*nn] by its discrete Fourier transform, if isign is input as 1; or replaces
 *   data[1..2*nn] by nn times its inverse discrete Fourier transform, if isign is input as -1.
 *   data is a complex array of length nn or, equivalently, a real array of length 2*nn. nn MUST
 *   be an integer power of 2 (this is not checked for!).
 *
 *    The real and imaginary parts of the zero frequency component F 0 are in data[1]
 * and data[2]; the smallest nonzero positive frequency has real and imaginary parts in
 * data[3] and data[4]; the smallest (in magnitude) nonzero negative frequency has
 * real and imaginary parts in data[2*nn-1] and data[2*nn]. Positive frequencies
 * increasing in magnitude are stored in the real-imaginary pairs data[5], data[6]
 * up to data[nn-1], data[nn]. Negative frequencies of increasing magnitude are
 * stored in data[2*nn-3], data[2*nn-2] down to data[nn+3], data[nn+4].
 * Finally, the pair data[nn+1], data[nn+2] contain the real and imaginary parts of
 * the one aliased point that contains the most positive and the most negative frequency.
 * You should try to develop a familiarity with this storage arrangement of complex
 * spectra, also shown in Figure 12.2.2, since it is the practical standard.
 * This is a good place to remind you that you can also use a routine like four1
 * without modification even if your input data array is zero-offset, that is has the range
 * data[0..2*nn-1]. In this case, simply decrement the pointer to data by one when
 * four1 is invoked, e.g., four1(data-1,1024,1);. The real part of f0 will now be
 * returned in data[0], the imaginary part in data[1], and so on.
 */
  // Note:
  // Data before (signal): N complex data points, array of length 2N+1.
  // [ 0      1  2      3  4      5  6                      2N-3 2N-2     2N-1 2N  ]
  // [ ?    (f1,f1)    (f2,f2)   (f3,f3)    ...    ...     (fN-1,fN-1)    (fN,2N)  ]
  // Data after (frequency):
  // [ 0   1  2       3  4                N+1  N+2       N+1       N+2       N+3         N+4             2N-1  2N  ]
  // [ ?  (F0,F0)    (F1,F1)    ...     (FN/2,FN/2)    (-FN/2 -1,-FN/2 -1)   (-FN/2 -2,-FN/2 -2)    ...  (-F1,-F1) ] 
  //                                   (-FN+1,-FN/2)
  // 
  //        *           *                    *                                                              *
  //    zero freq    positive          positive/negative                                                 negative     
  //               smallest freq         largest freq                                                  smallest freq                                     
#include <math.h>
#define SWAP(a,b) tempr=(a);(a)=(b);(b)=tempr


void fullfourier(double data[], unsigned long nn, int isign)

{
   unsigned long n,mmax,m,j,istep,i;
   double wtemp,wr,wpr,wpi,wi,theta; //Double precision for the trigonometric recurrences
   double tempr,tempi;
   n=nn << 1;
   j=1;
   for (i=1;i<n;i+=2)
     { //This is the bit-reversal section of the routine. 
	if (j > i) 
	  {
	     SWAP(data[j],data[i]); //Exchange the two complex numbers.
	     SWAP(data[j+1],data[i+1]);
	  }
	m=nn;
	while (m >= 2 && j > m) 
	  {
	     j -= m;
	     m >>= 1;
	  }
	j += m;
     }
   //Here begins the Danielson-Lanczos section of the routine.
   mmax=2;
   while (n > mmax) 
     { //Outer loop executed log2 nn times.
	istep=mmax << 1;
	theta=isign*(6.28318530717959/mmax); //Initialize the trigonometric recurrence.
	wtemp=sin(0.5*theta);
	wpr = -2.0*wtemp*wtemp;
	wpi=sin(theta);
	wr=1.0;
	wi=0.0;
	for (m=1;m<mmax;m+=2) 
	  { //Here are the two nested inner loops.
	     for (i=m;i<=n;i+=istep)
	       {
		  j=i+mmax; //This is the Danielson-Lanczos formula:
		  tempr=wr*data[j]-wi*data[j+1];
		  tempi=wr*data[j+1]+wi*data[j];
		  data[j]=data[i]-tempr;
		  data[j+1]=data[i+1]-tempi;
		  data[i] += tempr;
		  data[i+1] += tempi;
	       }
	     wr=(wtemp=wr)*wpr-wi*wpi+wr; //Trigonometric recurrence.
	     wi=wi*wpr+wtemp*wpi+wi;
	  }
	mmax=istep;
     }
};	
